"""
Viewsets have the advantage of abstracting away the work of defining routes explicitly,
but routers have some limits. They can only go to a certain depth in producing urls.

For instance, let's imagine a simple application, where you have Bands and Albums.
In case you wish to list all Albums of a Band, you could make a request to an enpoint
like `/bands/<band_id>/albums/`. That's totally possible with routers. But what if you
want a detail route for an Album of a Band? A route like `/bands/<band_id>/albums/<album_id>/`
would make sense, right? But routers aren't able to go to such an extent. And you could
totally imagine bigger urls in real, bigger applications.

So defining our routes manually gives us a lot more control. Everything comes with a tradeoff
though. When manually defining routes for generic views, you can easily assign each view class
to their routes, using the `as_view` method. But viewsets are different. One viewset class can
be assigned to more than one route. So for that to work, you've gotta do something like [this](https://www.django-rest-framework.org/tutorial/6-viewsets-and-routers/#binding-viewsets-to-urls-explicitly).

In order to simplify things, and abstract away some boiler plate code, this module provides the
standard viewset actions mapped to their corresponding http method. Of course, you may have additional
actions, customized according to your own needs. In this case, you can config them on your own. But
the standard ones are all set here.

Usage example:

---

```python
# urls.py

from django.urls import path
from dj_drf_utils.action_patterns import STANDARD_DETAIL_PATTERN, STANDARD_PATTERN

from . import views

cinema_view = views.CinemaViewSet.as_view(STANDARD_PATTERN)
cinema_detail_view = views.CinemaViewSet.as_view(STANDARD_DETAIL_PATTERN)

urlpatterns = [
    path("", cinema_view),
    path("<cinema_id>/", cinema_detail_view),
]
```

---

But routers are still so cool and so simple to use. So a very good alternative is [drf-nested-routers](https://github.com/alanjds/drf-nested-routers).
It really makes it easier to deal with all of this. The `drf-nested-routers` library is designed to
solve exactly this problem, and even more.
"""

LIST_PATTERN = {"get": "list"}
CREATE_PATTERN = {"post": "create"}

RETRIEVE_PATTERN = {"get": "retrieve"}
UPDATE_PATTERN = {"put": "update"}
PARTIAL_UPDATE_PATTERN = {"patch": "partial_update"}
DESTROY_PATTERN = {"delete": "destroy"}

STANDARD_PATTERN = {**LIST_PATTERN, **CREATE_PATTERN}
STANDARD_DETAIL_PATTERN = {
    **RETRIEVE_PATTERN,
    **UPDATE_PATTERN,
    **PARTIAL_UPDATE_PATTERN,
    **DESTROY_PATTERN,
}

RETRIEVE_UPDATE_PATTERN = {
    **RETRIEVE_PATTERN,
    **UPDATE_PATTERN,
    **PARTIAL_UPDATE_PATTERN,
}
RETRIEVE_DESTROY_PATTERN = {**RETRIEVE_PATTERN, **DESTROY_PATTERN}
UPDATE_DESTROY_PATTERN = {**UPDATE_PATTERN, **DESTROY_PATTERN}



"""
This module provides a custom user manager as a shortcut for whoever wants to customize
django's authentication system to use a different field instead of username for login.
It can be really anything, like email, phone, cpf, etc.
"""

from django.contrib.auth.models import BaseUserManager
from django.core.exceptions import FieldDoesNotExist, ValidationError


class CustomUserManager(BaseUserManager):
    """
    A custom user manager that inherits from `django.contrib.auth.models.BaseUserManager`.
    Its purpouse in life is mainly to provide an easy and simple way to implement a login
    and register system that expects another field instead of `username`.

    But what if you desired to customize your users in a way that other info is also required
    for user creation? No problem, this class is highly customizable.

    Instead of having to override the `create` and `create_superuser` methods of `BaseUserManager`,
    you can inherit from `CustomUserManager` and then simply set some class properties at your will.
    They work as follows:

    - `auth_field_name` -> Defaults to `"email"`. Defines what is the name of the field that
    should be used for login (besides password, of course). Note that this field **must**
    exist in your user model, **having a unique constraint**.
    - `user_is_staff` -> Defaults to `False`. Defines the starting staff status of newly
    created users
    - `user_start_active` -> Defaults to `True`. Defines if a user account should start in
    active state. In cases where users have to confirm their account in some way before getting
    access, you may wish to set this property to `False`
    - `super_is_staff` -> Defaults to `True`. Defines the starting staff status of newly
    created superusers
    - `super_start_active` -> Defaults to `True`. Defines if a superuser account should start in
    active state. Usually you'll want this value to be `True`, but you're totally free to change
    it, depending on your needs.
    - `required_fields` -> Defaults to `[]`. It should be a `list[str]`. This property defines
    which fields are required to be provided upon user creation, besides `self.auth_field_name` and
    `password`. The fields `is_staff`, `is_superuser` and `is_active` should also not be present in
    this list. It is worth noting that **all fields defined here, must also be defined in your user model**.
    Otherwise, a `ValidationError` is raised.

    Below is an example of how you may customize the behaviour of this class:

    ---

    ```python

    # managers.py

    from dj_drf_utils.managers import CustomUserManager

    class MyOwnUserManager(CustomUserManager):
        user_start_active = False
        required_fields = ["first_name", "last_name"]
    ```

    ---

    In order to implement a login with email feature, for instance, you have to make some minor
    changes to your user model. Below are some settings that may come in handy for you to define
    in your model:

    ---

    ```python

    # models.py

    from .managers import MyOwnUserManager
    from django.db import models
    from django.contrib.auth.models import AbstractUser

    class MyUser(AbstractUser):
        email = models.EmailField(unique=True)

        username = None

        objects = MyOwnUserManager()

        USERNAME_FIELD = objects.auth_field_name
        REQUIRED_FIELDS = objects.required_fields
    ```

    ---

    The `email` property is defined as unique, since it's gonna be used for login (as per the `USERNAME_FIELD`
    property). The `objects` property may be either the standard `dj_drf_utils.managers.CustomUserManager`
    or your own manager that inherits from it. In the example above, we are using our own user manager,
    with some minor customizations. `REQUIRED_FIELDS` refer to the fields you are prompted when creating a
    superuser(it must not include the value defined for `USERNAME_FIELD` or "password"). Defining it to
    `objects.required_fields` prevents you from making mistakes and being redundant. Note that in the example
    above we are droping the `username` column, but that's not necessary if you still want to have a username
    in your user model.
    """

    auth_field_name: str = "email"

    user_is_staff: bool = False
    user_start_active: bool = True

    super_is_staff: bool = True
    super_start_active: bool = True

    required_fields: list[str] = []

    def create(self, password: str, **extra_fields):
        FORBIDDEN_FIELDS = [
            self.auth_field_name,
            "password",
            "is_staff",
            "is_superuser",
            "is_active",
        ]

        try:
            auth_field = extra_fields.pop(self.auth_field_name)
        except KeyError:
            raise ValidationError(f"Missing {self.auth_field_name} field")

        required_values = {self.auth_field_name: auth_field.lower()}

        for field in self.required_fields:
            if field in FORBIDDEN_FIELDS:
                raise ValidationError(
                    f"Field {field} should not be in self.required_fields"
                )
            try:
                self.model._meta.get_field(field)
                required_values[field] = extra_fields.pop(field)
            except FieldDoesNotExist:
                raise ValidationError(f"Field {field} does not exist in {self.model}")
            except KeyError:
                raise ValidationError(f"Missing {field} field")

        user = self.model(
            is_active=self.user_start_active,
            is_staff=self.user_is_staff,
            **extra_fields,
            **required_values,
        )
        user.set_password(password)
        user.save(using=self._db)

        return user

    def create_superuser(self, password: str, **extra_fields):
        user = self.create(password, **extra_fields)

        user.is_superuser = True
        user.is_active = self.super_start_active
        user.is_staff = self.super_is_staff

        user.save(using=self._db)

        return user



"""
This module provides useful mixins to be used in Django Rest Framework **generic views** and **viewsets**.
"""

from rest_framework.serializers import Serializer


class SerializerByMethodMixin:
    """
    This mixin overrides the `get_serializer_class` method of generic views. It's
    purpose is to dinamically define which serializer to use, depending on the request
    method. For this to be possible, a new class property should be set, it is:

    - `method_serializers` -> It should be a dictionary having it's keys with the names
    of http methods and values as the serializer classes corresponding to each method.
    If the request method does not match any of the dict keys, it will return the value
    of `self.serializer_class`.

    Below is an example:

    ---

    ```python

    # views.py

    class MyBeautifulGenericView(SerializerByMethodMixin, ListCreateAPIView):
        queryset = MyWonderfulModel.objects.all()
        serializer_class = MyDefaultSerializer
        method_serializers = {
            "GET": MySerialzerToUseInGetRequests,
        }
    ```
    """

    method_serializers: dict[str, Serializer] = None

    def get_serializer_class(self):
        return self.method_serializers.get(self.request.method, self.serializer_class)


class SerializerByActionMixin:
    """
    This mixin overrides the `get_serializer_class` method of viewsets. It's
    purpose is to dinamically define which serializer to use, depending on the viewset
    action. For this to be possible, a new class property should be set, it is:

    - `action_serializers` -> It should be a dictionary having it's keys with the names
    of viewset actions and values as the serializer classes corresponding to each action.
    If the viewset action does not match any of the dict keys, it will return the value
    of `self.serializer_class`.

    Below is an example:

    ---

    ```python

    # views.py

    class MyBeautifulViewSet(SerializerByActionMixin, ModelViewSet):
        queryset = MyWonderfulModel.objects.all()
        serializer_class = MyDefaultSerializer
        action_serializers = {
            "create": MySerializerToUseInCreateActions,
            "update": MySerialzerToUseInUpdateActions,
            "partial_update": MySerialzerToUseInPartialUpdateActions,
        }
    ```
    """

    action_serializers: dict[str, Serializer] = None

    def get_serializer_class(self):
        return self.action_serializers.get(self.action, self.serializer_class)


class SerializerByDetailActionsMixin:
    """
    This mixin overrides the `get_serializer_class` method of viewsets. It's
    purpose is to dinamically define which serializer to use, depending on the viewset
    action. If it is a detail action, that is, one of `retrieve`, `update`, `partial_update`
    and `destroy`, then `self.detail_serializer_class` will be returned. Else, the default
    `self.serializer_class` is used. For this to be possible, a new class property should
    be set, it is:

    - `detail_serializer_class` -> It's value should be a serializer class. This property defines
    which serializer to use in detail actions.

    Below is an example:

    ---

    ```python

    # views.py

    class MyBeautifulViewSet(SerializerByDetailActionsMixin, ModelViewSet):
        queryset = MyWonderfulModel.objects.all()
        serializer_class = MyDefaultSerializer
        detail_serializer_class = MyDetailSerializer
    ```
    """

    detail_serializer_class: Serializer = None

    def get_serializer_class(self):
        DETAIL_ACTIONS = ["retrieve", "update", "partial_update", "destroy"]

        if self.action in DETAIL_ACTIONS:
            return self.detail_serializer_class

        return self.serializer_class


class SerializerBySafeActionsMixin:
    """
    This mixin overrides the `get_serializer_class` method of viewsets. It's
    purpose is to dinamically define which serializer to use, depending on the viewset
    action. If it is a _safe action_, then `self.safe_serializer_class` will be returned.
    Else, the default `self.serializer_class` is returned. A safe action is an action
    listed in the `safe_actions` class property. For this to be possible, a new class
    property should be set, it is:

    - `safe_serializer_class` -> Its value should be a serializer class. This property defines
    which serializer to use in safe actions.

    You can totally customize what is a "safe action". For that, you could change the value
    of `self.safe_actions`.

    - `safe_actions` -> It should be a `list[str]`, which each item representing a viewset action,
    considered safe for that viewset. The default value is `["list", "retrieve"]`

    Below is an example:

    ---

    ```python

    # views.py

    class MyBeautifulViewSet(SerializerBySafeActionsMixin, ModelViewSet):
        queryset = MyWonderfulModel.objects.all()
        serializer_class = MyDefaultSerializer
        safe_serializer_class = MySafeSerializer
    ```
    """

    safe_actions: list[str] = ["list", "retrieve"]
    safe_serializer_class: Serializer = None

    def get_serializer_class(self):
        if self.action in self.safe_actions:
            return self.safe_serializer_class

        return self.serializer_class



"""
This module provides three useful functions. Two of them are a more powerful and versatille version
of `get_object_or_404` and `get_list_or_404`, and the other is a handy shortcut.
"""

from django.core.exceptions import FieldError, ValidationError
from django.db.models import Model
from django.http import Http404
from rest_framework.exceptions import APIException


def get_object_or_error(klass: Model, exception: APIException = Http404, **kwargs):
    """
    Almost the same as `django.shortcuts.get_object_or_404`, but can raise any
    custom error class you want, allowing you to return more precise error messages.
    Another advantage of using this helper function, is that it prevents your application
    from crashing. For instance, in case you want to get an object by it's primary key, and
    it is of type `uuid`, but another data type is provided in the url, it will not crash,
    unlike the standard `get_object_or_404`. It expects the following arguments:

    - `klass` -> The model that will be used for the query
    - `exception` -> An error class inheriting from `rest_framework.exceptions.APIException`.
    If no `exception` is provided, then the standard `django.http.Http404` class is used.
    - `**kwargs` -> Keyword arguments representing all fields that should be used for the
    search, as many as you please.

    For instance, in case you want to get a `Room` of a `Cinema`:

    ---

    ```python

    # exceptions.py

    class CinemaNotFoundError(APIException):
        status_code = status.HTTP_404_NOT_FOUND
        default_detail = "Cinema not found"


    class RoomNotFoundError(APIException):
        status_code = status.HTTP_404_NOT_FOUND
        default_detail = "Room not found in this cinema"

    ```

    ---

    ```python

    # request endpoint

    "/cinemas/<cinema_id>/rooms/<room_id>/"

    ```

    ---

    ```python

    # views.py

    cinema = get_object_or_error(Cinema, CinemaNotFoundError, pk=self.kwargs['cinema_id'])
    room = get_object_or_error(Room, RoomNotFoundError, pk=self.kwargs['room_id'], cinema=cinema)

    ```

    ---

    Note that in case a room id is valid, but the cinema id is not, an appropriated message will be
    returned. In case you would use `get_object_or_404`, you would get just a `"Not found."`. Having
    more than one lookup field, `get_object_or_error` makes much clearer what is the problem.

    I highly encorage you to have a quick look at the source code, it's quite a simple concept.
    """
    try:
        return klass.objects.get(**kwargs)
    except (klass.DoesNotExist, ValidationError, ValueError, FieldError):
        raise exception


def get_list_or_error(
    klass: Model,
    error_klass: APIException = Http404,
    accept_empty: bool = False,
    **kwargs
):
    """
    Almost the same as `django.shortcuts.get_list_or_404`, but can raise any
    custom error class you want, allowing you to return more precise error messages.
    Another advantage of using this helper function, is that it prevents your application
    from crashing. For instance, in case you want to get a list, filtering it by some foreign
    key field, which is of type `uuid`, but another data type is provided in the url, it will
    not crash, unlike the standard `get_list_or_404`. Also, this function gives you the possiblity
    of not raising an exception when no values are found, so you could just return an empty list.
    It expects the following arguments:

    - `klass` -> The model that will be used for the query
    - `exception` -> An error class inheriting from `rest_framework.exceptions.APIException`.
    If no `exception` is provided, then the standard `django.http.Http404` class is used.
    - `accept_empty` -> A boolean argument, which defaults to `False`. When provided, determines
    if an empty result is acceptable or if it should raise `exception`.
    - `**kwargs` -> Keyword arguments representing all fields that should be used for the
    search, as many as you please.

    For instance, in case you want to list all `MovieSession`s of a `Room` in a `Cinema`:

    ---

    ```python

    # exceptions.py

    class NoMovieSessionsError(APIException):
        status_code = status.HTTP_404_NOT_FOUND
        default_detail = "This room has no scheduled movie sessions"

    ```

    ---

    ```python

    # request endpoint

    "/cinemas/<cinema_id>/rooms/<room_id>/movie-sessions/"

    ```

    ---

    ```python

    # views.py

    cinema = get_object_or_error(Cinema, CinemaNotFoundError, pk=self.kwargs['cinema_id'])
    room = get_object_or_error(Room, RoomNotFoundError, pk=self.kwargs['room_id'], cinema=cinema)
    movie_sessions = get_list_or_error(MovieSession, NoMovieSessionsError, room=room)

    ```

    ---

    I highly encorage you to have a quick look at the source code, it's quite a simple concept.
    """
    try:
        result = klass.objects.filter(**kwargs)
    except (ValidationError, ValueError, FieldError):
        raise error_klass

    if result or accept_empty:
        return result

    raise error_klass


def set_and_destroy(
    klass: Model, attr: str, value: list[Model], related_klass: Model, **kwargs
):
    """
    This function basically sets a new list of values in of a foreign key field and erases any
    previous values that were related to `klass`. For it to work, **you must set `null=True`
    in your model**, otherwise, the values will not be subsitituted, they will only be added.
    It accepts the following parameters:

    - `klass` -> The model on the side `1` of a `1:N` relationship, the owner of the relation,
    in which the new values will be set
    - `attr` -> A string version of the attribute corresponding to the `related_name` value
    in the foreign key field
    - `value` -> A list (or any other iterable), containing new created instances of `related_klass`
    - `related_klass` -> The model on the side `N` of a `1:N` relationship, the one having the foreign
    key field
    - `**kwargs` -> Keyword arguments used in a filter to determine which objects should be destroyed.
    It could be really anything, but usually you will want it to be something like `klass=None`, so
    that all objects that are no part of the relationship anymore can be descarded.

    For instance, a `Movie` may have many `Video`s related to it, like teasers and trailers. In case
    you want to update a `Movie`, reseting its `Video`s:

    ---

    ```python

    # models.py

    class Movie(models.Model):
        ...


    class Video(models.Model):
        id = models.UUIDField(primary_key=True, editable=False, default=uuid4)
        title = models.CharField(max_length=127)
        url = models.URLField()

        movie = models.ForeignKey(Movie, on_delete=models.CASCADE, related_name="videos", null=True)

    ```

    ---

    ```python

    # serializers.py

    class MovieSerializer(serializers.ModelSerializer):
        ...

        def update(self, instance: Movie, validated_data: dict):
        ...

        videos_data = validated_data.pop("videos", None)
        if videos_data:
            videos = [
                Video.get_or_create(**video, movie=instance)[0]
                for video in videos_data
            ]
            set_and_destroy(
                klass=instance,
                attr="videos",
                value=videos,
                related_klass=Video,
                movie=None,
            )

    ```

    ---

    In the example above, we are first getting or creating video instances, in order to reuse the ones
    passed in the body of the request that may already be in our db. Each video can only be related to
    one movie, since it doesn't make sense that two movies have the same trailer or teaser. So when
    assigning this new list of videos to a movie, the `set_and_destroy` function safely deletes all
    videos having their `movie` foreign key equal to `None`.

    I highly encorage you to have a quick look at the source code, it's quite a simple concept.
    """
    getattr(klass, attr).set(value)
    related_klass.objects.filter(**kwargs).delete()



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="dj_drf_utils",
    version="1.1.3",
    description="Useful functions and classes for Django and Django Rest Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cayo-rodrigues/django-utils",
    author="Cayo Rodrigues",
    author_email="cayo.rodrigues1914@gmail.com",
    license="MIT",
    keywords="django utils serializers generic views viewsets mixins email login model manager custom action shortcut error simple rest framework dj drf",
    packages=find_packages(include=["dj_drf_utils", "dj_drf_utils.*"]),
    install_requires=["djangorestframework"],
    python_requires=">=3.8",
)
